/*
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// ******************** Time section ********************

// Time definition when you first connect to power
int getStartMinutes(){
  int startHours =   16;
  int startMinutes = 01; 
  
  return startHours*60 + startMinutes;
}

// Get the number of times Arduino rolled over the milliseconds (max: 34359737)
int millisRollover() {
  static int numRollovers=0; // variable that permanently holds the number of rollovers since startup
  static boolean readyToRoll = false; // tracks whether we've made it halfway to rollover
  unsigned long now = millis(); // the time right now
  unsigned long halfwayMillis = 17179868; // this is halfway to the max millis value (17179868)

  if (now > halfwayMillis) { // as long as the value is greater than halfway to the max
    readyToRoll = true; // you are ready to roll over
  }

  if (readyToRoll == true && now < halfwayMillis) {
    // if we've previously made it to halfway
    // and the current millis() value is now _less_ than the halfway mark
    // then we have rolled over
    numRollovers = numRollovers++; // add one to the count the number of rollovers
    readyToRoll = false; // we're no longer past halfway
  } 
  return numRollovers;
}

long getElapsedMinutes(){
  return getStartMinutes() + (millis()/1000 + millisRollover()*3436)/60;
}

// Light state: On -> Blink -> Off
// light = [red,yellow,green]

// State ON until minutes:
int onTilMinutes[] =      {10, 5,  2};
// State LOBLINK until minutes:
int loBlinkTilMinutes[] = {8,  4,  1};
// State HIBLINK until minutes:
int hiBlinkTilMinutes[] = {6,  3, -1};

// ******************** Light section ********************

// Pin definitions
int lightPins[] = {9, 10, 11};
// Fade directions
boolean lightFadeDirection[] = {true, true, true};
// Light brightness
int lightBrightness[] = {0, 64, 128};

// Get light function in terms of remaining minutes
// light = 0 red 1 yellow 2 green
void getLightFunction(int lightColor, boolean &lightOn, boolean &loLightBlink, boolean &hiLightBlink, boolean &lightPrecog, int firstBusMinutes, int secondBusMinutes){
  if(firstBusMinutes > onTilMinutes[lightColor])
    if (lightColor == 0 || firstBusMinutes <= hiBlinkTilMinutes[lightColor-1])
      lightOn = true;
    else{
      lightOn = false;
      loLightBlink = false;
      hiLightBlink = false;
      lightPrecog = false;
    }
  else{
    lightOn = false;
    
    if(firstBusMinutes > hiBlinkTilMinutes[lightColor]){
      if (firstBusMinutes <= loBlinkTilMinutes[lightColor]){
        hiLightBlink = true;
        loLightBlink = false;
      }else if (firstBusMinutes <= onTilMinutes[lightColor]){
        hiLightBlink = false;
        loLightBlink = true;
      }else{
        loLightBlink = false;
        hiLightBlink = false;
        lightPrecog = false;
      }
    }else{
      loLightBlink = false;
      hiLightBlink = false;
      // If the green light is blinking
      if(lightColor != 2 && firstBusMinutes <= onTilMinutes[2] && firstBusMinutes > hiBlinkTilMinutes[2]){
        // If this light would be on or blink after that
        if(secondBusMinutes > hiBlinkTilMinutes[lightColor] && (lightColor == 0 || (secondBusMinutes <= hiBlinkTilMinutes[0])))
          lightPrecog = true;
        else
          lightPrecog = false;
      } else
        lightPrecog = false;
    }
  }
}

// Setup the brightness, depending on the blinking step.
void lightBlinkSetup(boolean &dir, int amount, int &brightness){
  if(dir)
    brightness = brightness + amount;
  else
    brightness = brightness - amount;
  
  if(brightness >= 255){
    brightness = 255;
    dir = !dir;
  } else if(brightness <= 0){
    brightness = 0;
    dir = !dir;
  }
}

// Update the lights status
void updateLights(int firstBusMinutes, int secondBusMinutes){
  for(int color=0; color<3; color++){
    boolean lightOn;
    boolean loLightBlink;
    boolean hiLightBlink;
    boolean lightPrecog;
    int brightness = 0;
    
    getLightFunction(color, lightOn, loLightBlink, hiLightBlink, lightPrecog, firstBusMinutes, secondBusMinutes);
    
    if(lightOn)
      brightness = 255;
    else if (loLightBlink || hiLightBlink){
      int blinkSpeed;
      
      if(loLightBlink)
        blinkSpeed = 10;
      else
        blinkSpeed = 40;
        
      lightBlinkSetup(lightFadeDirection[color], blinkSpeed, lightBrightness[color]);
      brightness = lightBrightness[color];
    }else if(lightPrecog)
      brightness = 5;
      
    analogWrite(lightPins[color], brightness);
  }
}

// A visible ntroduction, to show an eventual power outage and time reset
void lightIntro(){
  int amount = 5;
  int amountDirection = true;
  int amountCount = 0;
  
  boolean exit = false;
  
  while(!exit){
    amountCount++;
    if(amountCount == 20){
      
      if(amountDirection)
        amount = amount + 1;
      else
        amount = amount - 1;
        
      if(amount == 5)
        amountDirection = true;
      else if(amount == 20)
        amountDirection = false;

      amountCount = 0;
    }
    
    for (int i=0; i<3; i++){
      lightBlinkSetup(lightFadeDirection[i], amount, lightBrightness[i]);
      analogWrite(lightPins[i], lightBrightness[i]);
    }
      
    delay(30);
      
    if(isButtonPressed())
      exit = true;
    
    // Update the rollover count  
    millisRollover();
  }
}

// ******************** Bus section ********************

// Bus definitions
int bus177[] = {13,263,293,323,343,367,377,387,397,407,412,417,422,427,432,437,442,447,452,457,462,467,472,482,492,502,512,522,532,542,552,562,572,582,592,602,612,622,632,642,652,662,672,682,692,702,712,722,732,742,752,762,772,782,792,802,812,822,832,842,852,862,872,882,892,902,912,922,932,942,952,962,972,982,992,1002,1012,1022,1032,1042,1052,1062,1072,1082,1092,1102,1112,1122,1132,1142,1152,1172,1192,1213,1233,1253,1293,1333,1373,1413};
int bus177len = 100;
int bus277[] = {35,255,275,295,315,335,355,365,375,385,395,405,410,415,420,425,430,435,440,445,450,455,460,465,470,475,480,485,490,495,500,505,510,520,530,540,550,560,570,580,590,600,610,620,630,640,650,660,670,680,690,700,710,720,730,740,750,760,770,780,790,800,810,820,830,840,850,860,870,880,890,900,910,920,930,940,950,960,970,980,990,1000,1010,1020,1030,1040,1050,1060,1070,1080,1090,1100,1110,1120,1130,1140,1150,1160,1170,1175,1185,1195,1205,1215,1225,1235,1245,1265,1285,1305,1325,1345,1365,1395,1435};
int bus277len = 115;
int bus617[] = {62,122,182,242};
int bus617len = 4;

int timeToStation = 6;

// Update the minimums with a trickery to switch from first to second
void updateMins(long &minFirst, long &minSecond, long &curFirst, long &curSecond){
  long discarded;
  
  if(minFirst < curFirst) // We discard curFirst
    discarded = curFirst;
  else
    discarded = minFirst;
    
  // See if the discarded is now a minimum second...  
  minSecond = min(discarded, minSecond);
  
  //Update minFirst
  minFirst = min(minFirst, curFirst);
  // Update minSecond
  minSecond = min(minSecond, curSecond);
}

void getTimesToBus(int busTimes[], int busTimesLen, long &firstBus, long &secondBus, long time){
   int indexFirst = 0;
   int indexSecond = 0;
   while ( time > busTimes[indexFirst] && indexFirst < busTimesLen ){
     indexFirst++;
   }

  if ( indexFirst == busTimesLen )
    indexFirst = 0;

  indexSecond = (indexFirst + 1) % busTimesLen;
  
  firstBus = busTimes[indexFirst];
  if(time <= firstBus)
    firstBus = firstBus - time;
  else
    firstBus = firstBus + 60*24 - time;
    
  secondBus = busTimes[indexSecond];
  if(time <= secondBus)
    secondBus = secondBus - time;
  else
    secondBus = secondBus + 60*24 - time;
}

double getTimesToBusStation(long &firstBus, long &secondBus, long time){
  long minFirst;
  long minSecond;

  long curFirst;
  long curSecond;
  
  time = time + timeToStation;
  
  getTimesToBus(bus177, bus177len, curFirst, curSecond, time);
  
  minFirst = curFirst;
  minSecond = curSecond;
  
  getTimesToBus(bus277, bus277len, curFirst, curSecond, time);
  
  updateMins(minFirst, minSecond, curFirst, curSecond);
  
  getTimesToBus(bus617, bus617len, curFirst, curSecond, time);
  
  updateMins(minFirst, minSecond, curFirst, curSecond);
  
  firstBus = minFirst;
  secondBus = minSecond;
}

// ******************** Button section ********************
// Pin definition
int buttonPin = 2;
// De-noised button state
int buttonState = LOW;
// Noisy button state
int lastButtonState = LOW;
// Non-repetitive check
boolean hasBeenOff = true;
// Noisy last pressed time
long lastDebounceTime = 0;
// De-noising delay
long debounceDelay = 50;

// Checks and updates the button state.
void checkButton(){
  int reading = digitalRead(buttonPin);
    
  if (reading != lastButtonState) {
    lastDebounceTime = millis();
  } 
    
  if ((millis() - lastDebounceTime) > debounceDelay) {
    buttonState = reading;
    
    if(buttonState == LOW)
      hasBeenOff = true;
  }
    
  lastButtonState = reading;
}

// Returns "true" only one time when the button is pressed, with noise removal
boolean isButtonPressed(){
  checkButton();
  boolean result;
  
  if (hasBeenOff && buttonState == HIGH){
    hasBeenOff = false;
    return true;
  } else
    return false;
}

// Button timeout minutes
long buttonTimeoutMinutes = 30;
// Button last clicked 
long buttonLastMinutes = 0;
// Is timeout in order?
boolean buttonTimeoutOn = false;

// Starts the timeout
void startButtonTimeout(){
  buttonTimeoutOn = true;
  buttonLastMinutes = getElapsedMinutes();
}

// Stops the timeout
void stopButtonTimeout(){
  buttonTimeoutOn = false;
}

// Has the last timeout expired?
boolean hasTimeoutExpired(){
  boolean result = true;
  
  if(buttonTimeoutOn){
    long nowMinutes = getElapsedMinutes();
    if(((nowMinutes >= buttonLastMinutes) && (nowMinutes - buttonLastMinutes < buttonTimeoutMinutes))
       || ((nowMinutes < buttonLastMinutes) && (nowMinutes+60*24 - buttonLastMinutes < buttonTimeoutMinutes)))
      result = false;
  }
  
  if(result)
    stopButtonTimeout();
  return result;
}

// ******************** Arduino section ********************

void setup()  { 
  pinMode(lightPins[0], OUTPUT);
  pinMode(lightPins[1], OUTPUT);
  pinMode(lightPins[2], OUTPUT);
  pinMode(buttonPin, INPUT);
} 

void loop()  {
  lightIntro();
  ampelLoop();
}

void ampelLoop(){
  startButtonTimeout();
  
  while(true){
    if(hasTimeoutExpired())
      sleepMode();
    else
      ampelMode();
      
    delay(30);
  }
}

void sleepMode(){
  analogWrite(lightPins[0], 0);
  analogWrite(lightPins[1], 0);
  analogWrite(lightPins[2], 0);
  
  if(isButtonPressed())
    startButtonTimeout();
}

void ampelMode(){
  long dayMinutes = getElapsedMinutes();
  long firstBus;
  long secondBus;
  
  getTimesToBusStation(firstBus, secondBus, dayMinutes);
  
  updateLights(firstBus, secondBus);
  
  if(isButtonPressed())
    stopButtonTimeout();
}
